
from array import array
from datetime import timedelta, datetime
from io import BytesIO
import struct

from .util import BitFormat


class InvalidPacket(Exception):
    pass


class Packet:
    """Base class for the various datatypes. May be created from a file-like
    object or started from scratch.

    :param file: Source file to read from.
    :type file: file-like
    :param kwargs: Initial header values. Read from file if not provided.

    **Chapter 10 Header Attributes**

    - sync_pattern
    - channel_id
    - packet_length
    - data_length
    - header_version
    - sequence_number
    - secondary_header
    - ipts_source
    - rtc_sync_error
    - data_overflow_error
    - secondary_format
    - data_checksum
    - data_type
    - rtc
    - header_checksum

    **Secondary Header (if present)**

    - secondary_time
    - secondary_checksum

    **Other Members**

    .. py:attribute:: FORMAT
        :type: BitFormat

        Description of the chapter 10 header

    .. py:attribute:: SECONDARY_FORMAT
        :type: BitFormat

        Describes the secondary header

    .. py:attribute:: csdw_format
        :type: BitFormat
        :value: None

        Describes a datatype's channel specific data word (CSDW)
    """

    FORMAT = BitFormat('''
        u16 sync_pattern
        u16 channel_id
        u32 packet_length
        u32 data_length
        u8 header_version
        u8 sequence_number
        u1 secondary_header
        u1 ipts_source
        u1 rtc_sync_error
        u1 data_overflow_error
        u2 secondary_format
        u2 data_checksum
        u8 data_type
        u48 rtc
        u16 header_checksum''')

    SECONDARY_FORMAT = BitFormat('''
        u64 secondary_time
        p16 reserved
        u16 secondary_checksum''')

    csdw_format = BitFormat('u32 csdw')

    def __init__(self, file=None, parent=None, **kwargs):

        self.parent = parent

        # Set defaults
        for fmt in (self.FORMAT, self.SECONDARY_FORMAT, self.csdw_format):
            for name in fmt.names:
                setattr(self, name, 0)
        self.sync_pattern = 0xeb25

        self.__dict__.update(kwargs)
        self._messages = []
        self.buffer = None

        # If file is not given, start from scratch.
        if not file:
            self.clear()
            return

        # Read header if not done already.
        if not kwargs:
            raw_header = file.read(24)
            kwargs = self.FORMAT.unpack(raw_header)
            self.__dict__.update(kwargs)
        else:
            raw_header = self.FORMAT.pack(kwargs)

        # Compute checksum and update attributes with header values.
        self.header_sums = sum(array('H', raw_header)[:-1]) & 0xffff

        # Read the secondary header (if any).
        secondary = bytes()
        if self.secondary_header:
            secondary = file.read(12)
            self.secondary_sums = sum(array('H', secondary)[:-1]) & 0xffff
            self.__dict__.update(self.SECONDARY_FORMAT.unpack(file.read(12)))

        # Read into our own personal buffer.
        header_size = 24 + len(secondary)
        body = file.read(self.packet_length - header_size)
        self.buffer = BytesIO(raw_header + secondary + body)
        self.buffer.seek(header_size)

        self.validate()

        # Read channel specific data word (CSDW)
        self.__dict__.update(self.csdw_format.unpack(self.buffer.read(4)))

    def get_header_sum(self):
        raw_header = self.FORMAT.pack(self.__dict__)
        self.header_sums = sum(array('H', raw_header)[:-1]) & 0xffff
        return self.header_sums

    def get_time(self):
        """Return a timestamp for this packet. Depends on parent C10 object."""

        if self.parent and self.parent.last_time is not None:
            t = self.parent.last_time.time
            rtc = self.rtc - self.parent.last_time.rtc
            mask = 0xffffffffffff
            while rtc > mask:
                rtc -= mask
            return t + timedelta(seconds=rtc / 10_000_000)
        else:
            return datetime.now()

    def _read_messages(self):
        """Internal: ensure all messages are read in order to manipulate data
        in an internal list.
        """

        if self.buffer:
            header_size = 36 if self.secondary_header else 24
            self.buffer.seek(header_size + 4)
            self._messages = list(self)

    def __next__(self):
        """Return the next message until the end, then raise StopIteration."""

        if self.buffer is None:
            if not getattr(self, '_message_iterator', None):
                self._message_iterator = iter(self._messages)
            return next(self._message_iterator)

        if not getattr(self, 'Message', None):
            raise StopIteration

        try:
            msg = self.Message.from_packet(self)
            self._messages.append(msg)
            return msg
        except EOFError:
            self.buffer = None
            raise StopIteration

    def __iter__(self):
        if self.buffer:
            return self
        else:
            return iter(self._messages)

    def validate(self, silent=False):
        """Validate the packet using checksums and verifying fields. If silent
        = False raises InvalidPacket.
        """

        err = None
        if self.sync_pattern != 0xeb25:
            err = InvalidPacket('Incorrect sync pattern!')
        elif self.get_header_sum() != self.header_checksum:
            err = InvalidPacket('Header checksum mismatch!')
        elif self.secondary_header:
            if self.secondary_sums != self.secondary_checksum:
                err = InvalidPacket('Secondary header checksum mismatch!')
        if self.data_length > 524288:
            err = InvalidPacket(
                f'Data length {self.data_length} larger than allowed!')

        if err:
            if not silent:
                raise err
            return False
        return True

    def __len__(self):
        """Return length if we can find one, else raise TypeError."""

        if self.buffer is None and self._messages is not None:
            return len(self._messages)
        if hasattr(self, 'count'):
            return self.count
        elif hasattr(self, 'Message') and self.Message.length:
            msg_size = self.Message.length
            if getattr(self.Message, 'FORMAT', None):
                msg_size += self.Message.FORMAT.calcsize() // 8
            return (self.data_length - 4) // msg_size
        raise TypeError("object of type '%s' has no len()" % self.__class__)

    def _raw_body(self):
        """Returns the raw bytes of the packet body, including the CSDW."""

        # Pack messages into body
        body = b''.join(bytes(m) for m in self._messages)
        return self.csdw_format.pack(self.__dict__) + body

    def __bytes__(self):
        """Returns the entire packet as raw bytes."""

        self._read_messages()

        body = self._raw_body()
        self.data_length = len(body)

        # Add filler to maintain 32 bit alignment
        checksum_size = (0, 1, 2, 4)[self.data_checksum]
        while (checksum_size + len(body)) % 4:
            body += b'\0'

        # Compute the data checksum
        if self.data_checksum:
            fmt = 'xBHI'[self.data_checksum]
            checksum = sum(array(fmt, body))
            checksum &= (0, 0xff, 0xffff, 0xffffffff)[self.data_checksum]
            body += struct.pack(fmt, checksum)

        # Pack header (with updated checksum) and secondary header if needed.
        header_length = 36 if self.secondary_header else 24
        self.packet_length = header_length + len(body)
        self.header_checksum = self.get_header_sum()
        raw = self.FORMAT.pack(self.__dict__)
        if self.secondary_header:
            raw += self.SECONDARY_FORMAT.pack(self.__dict__)

        # Add body and filler
        return raw + body

    def __repr__(self):
        return '<{} {} bytes>'.format(
            self.__class__.__name__, getattr(self, 'packet_length', '?'))

    def append(self, *messages):
        """Add one or more messages to the packet."""

        self._read_messages()
        for message in messages:
            self._messages.append(message)

    def clear(self):
        """Remove all messages."""

        self._read_messages()
        self._messages = []

    def copy(self):
        """Duplicate this packet."""

        return self.__class__(**self.__dict__.copy())

    def remove(self, i):
        """Remove message at index 'i'"""

        self._read_messages()
        del self._messages[i]

    # Pickle compatability.
    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getstate__(self):
        self._read_messages()
        return {k: v for k, v in self.__dict__.items() if not callable(v)}


class Message:
    """The base class for packet message data. Subclasses define FORMAT,
    length, etc. to give the format and can be instantiated empty or from a
    packet using the from_packet method.

    :param bytes data: The binary data to be stored. May be empty (for
        instance, if FORMAT fully describes the format).
    :param Packet parent: The Packet object this message belongs to.
    :param kwargs: Arbitrary key-value pairs to add as attributes to the item
        instance. Used for IPH values.

    **Class Attributes**

    .. py:attribute:: FORMAT
        :type: BitFormat
        :value: None

        Describes the intra-packet header (IPH) or message format in general.

    .. py:attribute:: length
        :type: int
        :value: 0

        Byte size of message body. If not specified will look for a 'length'
        field in the IPH.

    **Instance Attributes**

    .. py:attribute:: parent
        :type: Packet or None

        The Packet object this message is attached to.

    .. py:attribute:: data
        :type: bytes

        Raw data not identified in FORMAT
    """

    FORMAT = None
    length = 0

    def __init__(self, data=b'', parent=None, **kwargs):
        if self.FORMAT:
            self.__dict__.update({name: 0 for name in self.FORMAT.names})
        self.__dict__.update(kwargs)
        self.parent = parent
        self.data = data

    def get_time(self):
        """Return a timestamp for this message based on parent packet."""

        if self.parent is None or not hasattr(self, 'ipts'):
            raise AttributeError(
                'Must have parent packet and IPTS to get time.')

        rtc = self.ipts - self.parent.rtc
        mask = 0xffffffffffff
        while rtc > mask:
            rtc -= mask
        return self.parent.get_time() + timedelta(seconds=rtc / 10_000_000)

    @classmethod
    def from_packet(cls, packet):
        """Helper method to read a message from packet."""

        # Exit if we've read all messages.
        if hasattr(packet, 'count') and packet.count == len(packet._messages):
            raise EOFError

        # Exit when we reach the end of the packet body
        end = packet.data_length + (packet.secondary_header and 36 or 24)
        if packet.buffer.tell() >= end:
            raise EOFError

        # Read and parse the IPH
        iph = {}
        if getattr(cls, 'FORMAT', None):
            iph_size = cls.FORMAT.calcsize() // 8
            raw = packet.buffer.read(iph_size)
            if len(raw) < iph_size:
                raise EOFError
            iph = cls.FORMAT.unpack(raw)

        # Read the message data and account for filler if length is odd.
        length = iph.get('length', cls.length)
        data = packet.buffer.read(length)
        if length % 2:
            packet.buffer.seek(1, 1)

        return packet.Message(data, parent=packet, **iph)

    def __bytes__(self):
        """Return bytes() containing any IPH and data."""

        raw = bytes()
        if self.FORMAT is not None:
            raw += self.FORMAT.pack(self.__dict__)
        raw += self.data
        if len(raw) % 2:
            raw += b'\x00'
        return raw




from datetime import datetime, timedelta

from .util import BitFormat
from .packet import Packet


class TimeF1(Packet):
    """Time Packets

    **Note:** Low level fields such as HSn, TSn, etc. are not documented here
    nor are they intended to be used directly. Use the .time attribute instead.

    .. py:attribute:: time_source

        * 0 - Internal (recorder clock)
        * 1 - External
        * 2 - Internal from RMM
        * 15 - None

    .. py:attribute:: time_format

        Indicates time data packet format

        * 0 - IRIG-B
        * 1 - IRIG-A
        * 2 - IRIG-G
        * 3 - Real-Time clock
        * 4 - UTC time from GPS
        * 5 - Native GPS time
        * 15 - None (payload invalid)

    .. py:attribute:: leap

        Indicates if this is a leap year.

    .. py:attribute:: date_format

        0 for IRIG day format, 1 for month and year format.

    .. py:attribute:: irig_source

        IRIG time source

        * 0 - TCG freewheeling (no/lost time source)
        * 1 - TCG freewheeling from .TIME command
        * 2 - TCG freewheeling from RMM time
        * 3 - TCG locked to external IRIG time signal
        * 4 - TCG locked to external GPS
        * 5 - TCG locked to external Network Time Protocol (NTP)
        * 6 - TCG locked to external Precision Time Protocol (PTP)
        * 7 - TCG locked to external embedded time from input track/channel\
such as PCM or 1553

    .. py:attribute:: time

        Python datetime object representing the packet payload.
    """

    csdw_format = BitFormat('''
        u4 time_format
        u4 time_source

        u4 irig_source

        p2
        u1 date_format
        u1 leap

        p16''')

    def __init__(self, *args, **kwargs):
        Packet.__init__(self, *args, **kwargs)

        self.data_format = '''
            u4 Hmn
            u4 Tmn
            u4 TSn
            u4 Sn

            u4 TMn
            u4 Mn
            u4 THn
            u4 Hn

            u4 TDn
            u4 Dn
            '''

        if not self.date_format:
            self.data_format += 'u8 HDn'
        else:
            self.data_format += '''
                p3
                u1 TOn
                u4 On

                u4 TYn
                u4 Yn
                p2
                u2 OYn
                u4 HYn
                '''

        self.data_format = BitFormat(self.data_format)

        if not self.buffer:
            for name in self.data_format.names:
                setattr(self, name, 0)
            self._initial_time = 0
            if not self.time:
                self.time = datetime.now()
            return

        raw = self.buffer.read(self.data_length - 4)
        self.__dict__.update(self.data_format.unpack(raw))

        ms = ((self.Hmn * 10) + self.Tmn)
        seconds = self.Sn + (self.TSn * 10)
        minutes = self.Mn + (self.TMn * 10)
        hours = self.Hn + (self.THn * 10)

        # IRIG day format
        if not self.date_format:
            day = self.Dn + (self.HDn * 100) + (self.TDn * 10)

            today = datetime.today()
            self.time = datetime(today.year, 1, 1) + timedelta(day - 1)

        # Month and Year Format
        else:
            month = self.On + (self.TOn * 10)
            day = self.Dn + (self.TDn * 10)
            year = self.Yn + (self.TYn * 10) + (self.HYn * 100) + (
                self.OYn * 1000)
            self.time = datetime(year, month, day)

        self.time = self.time.replace(
            microsecond=ms * 1000,
            second=seconds,
            minute=minutes,
            hour=hours,
            tzinfo=None)
        self._initial_time = self.time

    def _raw_body(self):
        ms = self.time.microsecond // 1000
        self.Hmn = ms // 100
        self.Tmn = (ms - (self.Hmn * 100)) // 10

        self.TSn = self.time.second // 10
        self.Sn = self.time.second - (self.TSn * 10)
        self.TMn = self.time.minute // 10
        self.Mn = self.time.minute - (self.TMn * 10)
        self.THn = self.time.hour // 10
        self.Hn = self.time.hour - (self.THn * 10)

        # Month and year format
        if self.date_format:
            year = self.time.year
            self.OYn = year // 1000
            year -= self.OYn * 1000
            self.HYn = year // 100
            year -= self.HYn * 100
            self.TYn = year // 10
            year -= self.TYn * 10
            self.Yn = year

        else:
            day = int(self.time.strftime('%j'))
            self.HDn = day // 100
            day -= self.HDn * 100
            self.TDn = day // 10
            day -= self.TDn * 10
            self.Dn = day

        body = self.data_format.pack(self.__dict__)
        if len(body) % 2:
            body += b'\0'

        # Add CSDW and body
        return self.csdw_format.pack(self.__dict__) + body

    def __bool__(self):
        return bool(self.time)




from io import BytesIO
import os

from .packet import Packet, InvalidPacket
from .util import Buffer
from .computer import ComputerF0, ComputerF1, ComputerF2, ComputerF3
from .pcm import PCMF1
from .time import TimeF1
from .ms1553 import MS1553F1, MS1553F2
from .analog import AnalogF1
from .discrete import DiscreteF1
from .message import MessageF0
from .arinc429 import ARINC429F0
from .video import VideoF0, VideoF1, VideoF2
from .image import ImageF0, ImageF1, ImageF2
from .uart import UARTF0
from .i1394 import I1394F0, I1394F1
from .parallel import ParallelF0
from .ethernet import EthernetF0, EthernetF1

__all__ = ('TYPES', 'C10')

# Top level data types.
TYPES = {
    0x00: ComputerF0,
    0x01: ComputerF1,  # TMATS
    0x02: ComputerF2,  # Event
    0x03: ComputerF3,  # Index
    0x09: PCMF1,
    0x11: TimeF1,
    0x19: MS1553F1,
    0x1A: MS1553F2,
    0x21: AnalogF1,
    0x29: DiscreteF1,
    0x30: MessageF0,
    0x38: ARINC429F0,
    0x40: VideoF0,
    0x41: VideoF1,
    0x42: VideoF2,
    0x48: ImageF0,
    0x49: ImageF1,
    0x4A: ImageF2,
    0x50: UARTF0,
    0x58: I1394F0,
    0x59: I1394F1,
    0x60: ParallelF0,
    0x68: EthernetF0,
    0x69: EthernetF1,
}


class C10(object):
    """A Chapter 10/11 parser.

    :param f: A file like object or file path to read from.
    :type f: file or str

    .. py:attribute:: last_time
        :type: time.TimeF1

        Most recently parsed (or generated) time packet.
    """

    def __init__(self, f):
        if isinstance(f, str):
            f = open(f, 'rb')
        self.file = Buffer(f)
        self.last_time = None

    @classmethod
    def from_string(cls, s):
        """Create a C10 object from a string or bytes."""

        return cls(BytesIO(s))

    def __next__(self):
        """Walk a chapter 10 file using python's iterator protocol and return
        a chapter10.packet.Packet object for each valid packet found.
        """

        while True:
            pos = self.file.tell()
            try:
                header = Packet.FORMAT.unpack(self.file.read(24))
                handler = TYPES.get(header['data_type'], None)
                if handler:
                    packet = handler(self.file, parent=self, **header)
                    if packet.data_type == 0x11:
                        self.last_time = packet
                    return packet

                raise NotImplementedError('Type %s not implemented' %
                                          hex(header['data_type']))
            except EOFError:
                raise StopIteration
            except InvalidPacket:
                self.file.seek(pos + 1)

    def __repr__(self):
        return '<C10: {}>'.format(os.path.basename(self.file.name))

    def __iter__(self):
        return self




from .packet import Packet
from .util import BitFormat


class ParallelF0(Packet):
    """
    .. py:attribute:: scan_number

        If present contains the scan number of the first scan (for DCRsi data)

    .. py:attribute:: type

        If between 0x10 and 0x80 contains the number of bits of recorded data
    """

    csdw_format = BitFormat('''
        u24 scan_number
        u8 type
    ''')




from .util import BitFormat
from . import packet

from bitstring import Bits


class AnalogF1(packet.Packet):
    """
    .. py:attribute:: mode

        Indicates packing and aligmnent mode.

        * 0 - Packed
        * 1 - Unpacked, lsb padded
        * 3 - Unpacked, msb padded

    .. py:attribute:: length

        Bit length of samples.

    .. py:attribute:: subchannel

        Subchannel ID

    .. py:attribute:: subchannel_count

        Number of subchannels (and CSDWs) in the packet.

    .. py:attribute:: factor

        Exponent of power of 2 sampling rate factor denominator.

    .. py:attribute:: same

        Indicates whether this CSDW applies to all subchannels.

    """

    csdw_format = BitFormat('''
        u6 length
        u2 mode

        u8 subchannel
        u8 subchannel_count

        p3
        u1 same
        u4 factor
    ''')

    class Message(packet.Message):
        def __repr__(self):
            return '<Analog Sample %s bits>' % len(self.length)

        @classmethod
        def from_packet(cls, packet):
            subchannel = packet._subchannel
            if packet.same:
                csdw = packet.subchannels[0]
            else:
                csdw = packet.subchannels[subchannel]

            # Find the sample size in bytes (16 bit aligned).
            length = csdw['length'] or 64  # Length 0 = 64 bits

            if packet.bit_offset > len(packet.data):
                raise EOFError

            data = packet.data[packet.bit_offset:packet.bit_offset + length]
            packet.bit_offset += length

            packet._subchannel += 1
            if packet._subchannel >= len(packet.subchannels):
                packet._subchannel = 0

            return packet.Message(data.bytes, parent=packet, **csdw)

        def __bytes__(self):
            return self.data

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        self._subchannel = 0

        if self.length == 0:
            self.length = 16

        keys = (
            'mode',
            'length',
            'subchannel',
            'subchannel_count',
            'factor',
            'same')
        self.subchannels = [{k: getattr(self, k) for k in keys}]

        # Read CSDWs for additional subchannels if applicable.
        if not self.same:
            for i in range(self.subchannel_count - 1):
                self.subchannels.append(
                    self.csdw_format.unpack(self.buffer.read(4)))

        # Get the raw bytes of data since samples are specified in bit length
        raw = self.buffer.read(
            self.data_length - (4 * len(self.subchannels)))
        self.data = Bits(raw)
        self.bit_offset = 0

    def _raw_body(self):
        raw = bytes()
        for channel in self.subchannels:
            raw += self.csdw_format.pack(channel)
        raw += b''.join(bytes(msg) for msg in self._messages)
        return raw




from .util import BitFormat
from . import packet


__all__ = ('VideoF0', 'VideoF1', 'VideoF2')


class Video(packet.Packet):
    """Generic video superclass."""

    class Message(packet.Message):
        length = 188

        def __repr__(self):
            return '<MPEG Frame %s bytes>' % len(self.data)

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        if self.iph:
            self.Message.FORMAT = BitFormat('u64 ipts')


class VideoF0(Video):
    """MPEG2/H.264

    .. py:attribute:: byte_alignment

        * 0 - Little endian
        * 1 - Big endian

    .. py:attribute:: payload

        * 0 - MPEG-2 MP @ ML
        * 1 - H.264 MP @ L2.1
        * 2 - H.264 MP @ L2.2
        * 3 - H.264 MP @ L3

    .. py:attribute:: key_length

        Indicates if key-length-value metadata is present

    .. py:attribute:: scr_rtc_sync

        Indicates if MPEG-2 SCR is RTC

    .. py:attribute:: iph
    .. py:attribute:: embedded_time

        Indicates time embedded in MPEG-2 stream.
    """

    csdw_format = BitFormat('''
        p23
        u1 byte_alignment
        u4 payload
        u1 key_length
        u1 scr_rtc_sync
        u1 iph
        u1 embedded_time''')

    class Message(Video.Message):
        """
        .. py:attribute:: ipts

            If indicated by iph flag in CSDW (see above)
        """


class VideoF1(Video):
    """ISO 13818-1 MPEG-2 bit stream

    .. py:attribute:: count
    .. py:attribute:: type

        * 0 - Transport data bit stream
        * 1 - Program data bit stream

    .. py:attribute:: mode

        * 0 - CBR stream
        * 1 - Variable bit rate stream

    .. py:attribute:: embedded_time

        Indicates time embedded in MPEG-2 stream.

    .. py:attribute:: encoding_profile

        * 0 - Simple profile @ main level
        * 1 - Main profile @ low level
        * 2 - Main profile @ main level
        * 3 - Main profile @ high-1440 level
        * 4 - Main profile @ high level
        * 5 - SNR profile @ low level
        * 6 - SNR profile @ main level
        * 7 - Spatial profile @ high-1440 level
        * 8 - High profile @ main level
        * 9 - High profile @ high-1440 level
        * 10 - High profile @ high level
        * 11 - 4:2:2 profile @ main level

    .. py:attribute:: iph
    .. py:attribute:: scr_rtc_sync

        Indicates if MPEG-2 SCR is RTC

    .. py:attribute:: key_length

        Indicates if key length value metadata is present.

    """

    csdw_format = BitFormat('''
        u12 count
        u1 type
        u1 mode
        u1 embedded_time
        u4 encoding_profile
        u1 iph
        u1 scr_rtc_sync
        u1 key_length
        p10''')

    class Message(Video.Message):
        """
        .. py:attribute:: ipts

            If indicated by iph flag in CSDW (see above)
        """


class VideoF2(Video):
    csdw_format = BitFormat('''
        u12 count
        u1 type
        u1 mode
        u1 embedded_time
        u4 encoding_profile
        u1 iph
        u1 scr_rtc_sync
        u1 key_length_value
        u4 encoding_level
        u1 audio_encoding_type
        p5''')

    class Message(Video.Message):
        """
        .. py:attribute:: ipts

            If indicated by iph flag in CSDW (see above)
        """




from .util import BitFormat
from . import packet


class ARINC429F0(packet.Packet):
    """
    .. py:attribute:: count

        Number of ARINC 429 words in packet body.
    """

    csdw_format = BitFormat('''
        u16 count
        p16''')

    class Message(packet.Message):
        """
        .. py:attribute:: gap_time

            Gap time from the beginning of the preceding bus word (regardless \
of bus) to the beginning of the current bus word in 0.1-us increments.
        .. py:attribute:: bus_speed

            * 0 - low-speed bus (12.5kHz)
            * 1 - high-speed bus (100kHz)

        .. py:attribute:: parity_error
        .. py:attribute:: format_error
        .. py:attribute:: bus

            ARINC bus number (0-255)
        """
        FORMAT = BitFormat('''
            u16 gap_time

            u1 format_error
            u1 parity_error
            u1 bus_speed
            p1

            u4 gap_upper

            u8 bus
        ''', '211')
        length = 4

        # TODO: there should be a neater way to do this with bitstruct, but I
        # haven't found it yet.
        def __init__(self, *args, **kwargs):
            packet.Message.__init__(self, *args, **kwargs)
            self.gap_time += (self.gap_upper << 16)

        def __bytes__(self):
            old_gap = self.gap_time
            self.gap_upper = self.gap_time >> 16
            self.gap_time &= 0xffff
            b = packet.Message.__bytes__(self)
            self.gap_time = old_gap
            return b

        def __repr__(self):
            return '<ARINC-429 Data Word>'




try:
    import cbitstruct as bitstruct
except ImportError:
    import bitstruct


class BitFormat:
    """Bitstruct wrapper that allows for compiling from readable format
    strings.

    :param str src: "struct" like format with newline separated pairs of <fmt>,
        <name> where fmt is a bitstruct format string (generally as
        "<bitsize>") and name is the attribute name to use.
    :param str byteswap: Optional bitstruct style byteswap description. See
        bitstruct docs
    :value byteswap: None

    .. py:attribute:: fmt_str
        :type: string

        The original format string.

    .. py:attribute:: names
        :type: list

        A list of the field names from the format string.
    """

    def __init__(self, src, byteswap=None):
        self.byteswap = byteswap
        self.fmt_str = ''
        self.names = []
        for line in src.strip().splitlines():
            line = line.strip().split()
            if not line:
                continue
            if len(line) == 1 or line[-1].lower().endswith('reserved'):
                fmt = line[0]
            else:
                fmt, name = line
                self.names.append(name)
            self.fmt_str += fmt

        # Default to little endian if we're not explicitly defining swapping
        if not byteswap and self.fmt_str[-1] not in '<>':
            self.fmt_str += '<'

        self._compiled = bitstruct.compile(self.fmt_str, names=self.names)

    def __getattr__(self, name, default=None):
        if name in ('byteswap', 'pack', 'unpack', 'raw', 'fmt_str', 'names'):
            return object.__getattr__(self, name, default)
        return getattr(self._compiled, name, default)

    def unpack(self, data):
        """Use compiled format to unpack data. Uses self.byteswap if provided.

        :param bytes data: Raw data to unpack
        :returns dict: Unpacked values based on compiled format.
        """

        if self.byteswap:
            data = bitstruct.byteswap(self.byteswap, data)
        self.raw = data
        return self._compiled.unpack(data)

    def pack(self, values):
        raw = self._compiled.pack(values)
        if self.byteswap:
            raw = bitstruct.byteswap(self.byteswap, raw)
        return raw


# TODO: is this needed any more?
class Buffer(object):
    """File wrapper that raises EOF on a short read."""

    def __init__(self, io):
        self.io = io
        self.tell = self.io.tell
        self.seek = self.io.seek

    def read(self, size=None):
        value = self.io.read(size)
        if size and len(value) != size:
            raise EOFError
        return value




from .util import BitFormat
from . import packet


class I1394F0(packet.Packet):
    """
    .. py:attribute:: count
    .. py:attribute:: sync

        Synchronization code

    .. py:attribute:: packet_body_type
    """

    csdw_format = BitFormat('''
        u16 count
        p9
        u4 sync
        u3 packet_body_type''')

    class Message(packet.Message):
        """
        .. py:attribute:: reset

            If packet type 0 (bus status) indicates if bus reset has occurred

        .. py:attribute:: ipts

            Present for packet type = 2 (general purpose)
        """

        def __repr__(self):
            return '<IEEE-1394 Transaction>'

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        # Bus Status
        if self.packet_body_type == 0:
            self.Message.FORMAT = BitFormat('''
                p31
                u1 reset''')

        # Data Streaming
        elif self.packet_body_type == 1:
            self.Message.length = self.packet.data_length - 4

        # General Purpose
        elif self.packet_body_type == 2:
            self.Message.length = (
                (self.packet.data_length - 4) / self.transaction_count) - 8
            self.Message.FORMAT = BitFormat('u64 ipts')


class I1394F1(packet.Packet):
    """IEEE 1394 Physical Layer

    .. py:attribute:: count
    """

    csdw_format = BitFormat('''
        u16 count
        p16''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: length

            Transfer length (bytes)

        .. py:attribute:: buffer_overflow
        .. py:attribute:: overflow_error

            * 0 - no overflow
            * 1 - transfer started correctly, but is longer than standard \
transfer length
            * 2 - previous transfer was a type 1 overflow, this transfer \
ended correctly
            * 3 - previous transfer was a type 1 overflow and this transfer \
did not end correctly
        .. py:attribute:: speed

            See chapter 10 standard

        .. py:attribute:: status
        """

        FORMAT = BitFormat('''
            u64 ipts
            u16 length
            p1
            u1 buffer_overflow
            u2 overflow_error
            u4 speed
            u8 status''')




from .util import BitFormat
from . import packet


class EthernetF0(packet.Packet):
    """Ethernet data

    .. py:attribute:: count
    .. py:attribute:: ttb

        Time tag bits:

        * 0 - First bit of the frame destination address
        * 1 - Last bit of the frame check sequence
        * 2 - First bit of the frame payload data
        * 3 - Last bit of the frame payload data

    .. py:attribute:: format

        Should be 0 (IEEE 802.3 MAC frame)
    """

    csdw_format = BitFormat('''
        u16 count
        p9
        u3 ttb
        u4 format''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: length

            Payload length (bytes)

        .. py:attribute:: data_length_error
        .. py:attribute:: data_crc_error
        .. py:attribute:: network_id
        .. py:attribute:: crc_error
        .. py:attribute:: frame_error
        .. py:attribute:: content

            * 0 - Full MAC frame
            * 1 - Payload only (14 bytes from the destination address)

        .. py:attribute:: ethernet_speed

            * 0 - Auto
            * 1 - 10 Mbps
            * 2 - 100 Mbps
            * 3 - 1 Gbps
            * 4 - 10 Gbps
        """

        def __repr__(self):
            return '<Ethernet frame %s bytes>' % len(self.data)

        # Note: bitfields may need to be listed in reverse of expected
        # order.
        FORMAT = BitFormat('''
            u64 ipts

            u16 length

            u8 network_id
            u1 crc_error
            u1 frame_error
            u2 content
            u4 ethernet_speed
        ''')

        def __init__(self, *args, **kwargs):
            packet.Message.__init__(self, *args, **kwargs)
            self.length_error = self.length >> 14 & 0b1
            self.data_error = self.length >> 15
            self.length &= 0b11111111111111

        def __bytes__(self):
            self.length |= (self.length_error << 14) | (self.data_error << 15)
            return packet.Message.__bytes__(self)


class EthernetF1(packet.Packet):
    """ARINC-664

    .. py:attribute:: count
    .. py:attribute:: iph_length

        Fixed at 28
    """

    csdw_format = BitFormat('''
        u16 count
        u16 iph_length''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: flags

            * 0 - Actual data
            * 1 - Simulated data

        .. py:attribute:: error
        .. py:attribute:: length

            Message length (bytes)

        .. py:attribute:: virtual_link

            Lower 16 bits of the destination MAC address

        .. py:attribute:: src_ip

            Source IP address from ARINC IP header

        .. py:attribute:: dst_ip

            Destination IP address from ARINC IP header

        .. py:attribute:: dst_port

            Destination port from the ARINC UDP header

        .. py:attribute:: src_port

            Source port from the ARINC UDP header
        """

        def __repr__(self):
            return '<ARINC-664 Frame %s bytes>' % len(self.data)

        FORMAT = BitFormat('''
            u64 ipts
            u8 flags
            u8 error
            u16 length
            u16 virtual_link
            p16
            u32 src_ip
            u32 dst_ip
            u16 dst_port
            u16 src_port''')




from .util import BitFormat
from . import packet


__all__ = ('ImageF0', 'ImageF1', 'ImageF2')


class ImageMessage:

    def __repr__(self):
        return '<Image Segment>'


class ImageF0(packet.Packet):
    """Image data

    .. py:attribute:: segment_length
    .. py:attribute:: iph
    .. py:attribute:: sum

        * 0 - Less than one complete image
        * 1 - One complete image
        * 2 - Multiple complete images
        * 3 - Multiple incomplete images

    .. py:attribute:: parts

        Indicates which piece[s] are of the frame are contained in the packet:
    """

    csdw_format = BitFormat('''
        u27 length
        u1 iph
        u3 sum
        u3 parts''')

    class Message(packet.Message, ImageMessage):
        """
        .. py:attribute:: ipts

            If IPH is true (see above), containts intra-packet timestamp
        """

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        if self.iph:
            self.Message.FORMAT = BitFormat('u64 ipts')


class ImageF1(packet.Packet):
    """Still imagery

    .. py:attribute:: format

        * 0 - MIL-STD-2500 National Imagery Transmission Format
        * 1 - JPEG File Interchange Format
        * 2 - JPEG 2000 (ISO/IEC 154444-1)
        * 3 - Portable Network Graphics Format (PNG)

    .. py:attribute:: iph
    .. py:attribute:: sum

        * 0 - Contains less than one complete image
        * 1 - Contains one complete image
        * 2 - Contains multiple complete images
        * 3 - Contains multiple incomplete messages

    .. py:attribute:: parts

        * 0 - Doesn't contain first or last segment of the image
        * 1 - Contains first segment of image
        * 2 - Contains multiple complete messages
        * 3 - Contains both first and last segment of image
    """

    csdw_format = BitFormat('''
        p23
        u4 format
        u1 iph
        u2 sum
        u2 parts''')

    class Message(packet.Message, ImageMessage):
        """
        .. py:attribute:: ipts

            If IPH is true (see above), containts intra-packet timestamp

        .. py:attribute:: length

            Length of image or segment (bytes)
        """

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)
        fmt = ''
        if self.iph:
            fmt = 'u64 ipts\n'
        self.Message.FORMAT = BitFormat(fmt + 'u32 length')


class ImageF2(packet.Packet):
    """Dynamic Imagery

    .. py:attribute:: format

        Refer to chapter 10 standard

    .. py:attribute:: iph
    .. py:attribute:: sum

        * 0 - Contains less than one complete image (segment)
        * 1 - Contains one complete image
        * 2 - Contains multiple complete images

    .. py:attribute:: parts

        * 0 - Doesn't contain first or last segment of the image
        * 1 - Contains first segment of image
        * 2 - Contains last segment of image
    """

    csdw_format = BitFormat('''
        p21
        u6 format
        u1 iph
        u2 sum
        u2 parts''')

    class Message(packet.Message, ImageMessage):
        """
        .. py:attribute:: ipts

            If IPH is true (see above), containts intra-packet timestamp

        .. py:attribute:: length

            Length of image or segment (bytes)
        """

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)
        fmt = ''
        if self.iph:
            fmt = 'u64 ipts\n'
        self.Message.FORMAT = BitFormat(fmt + 'u32 length')




from collections import OrderedDict

from .util import BitFormat, bitstruct
from . import packet


class ComputerF0(packet.Packet):
    """User-defined"""

    def __init__(self, *args, **kwargs):
        self.data = bytes()
        packet.Packet.__init__(self, *args, **kwargs)
        if self.buffer:
            self.data = self.buffer.read(self.data_length - 4)

    def _raw_body(self):
        return (b'\0' * 4) + self.data


class ComputerF1(packet.Packet):
    """Setup record (TMATS)

    Using dictionary lookup syntax returns an OrderedDict of values that start
    with the given string::

        >> tmats_packet['G']
        OrderedDict({'G\\COM': 'Comment'})

    .. py:attribute:: version

        Irig106 release:

        * 7 = 2007
        * 8 = 2009
        * 9 = 2010
        * 10 = 2013
        * 11 = 2015
        * 12 = 2017

    .. py:attribute:: configuration_change

        When streaming, indicates if the TMATS configuration has changed from
        the previous one.

    .. py:attribute:: format

        Indicates ASCII (0) or XML (1)
    """

    csdw_format = BitFormat('''
        u8 version
        u1 configuration_change
        u1 format
        p22''')

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)
        if self.buffer:
            self.data = self.buffer.read(self.data_length - 4)
        elif not getattr(self, 'data', None):
            self.data = kwargs.get('data', bytes())

    def _raw_body(self):
        csdw = self.csdw_format.pack(self.__dict__)
        if isinstance(self.data, str):
            return csdw + bytes(self.data, 'utf8')
        return csdw + self.data

    def __getitem__(self, key):
        key = bytearray(key, 'utf-8')
        d = OrderedDict()
        for line in self.data.splitlines():
            if not line.strip() or b':' not in line:
                continue
            line = line.strip()[:-1].split(b':', 1)  # Strip the semicolon.
            if line[0].startswith(key):
                d[line[0]] = line[1]
        return d


class ComputerF2(packet.Packet):
    """Recording Event

    .. py:attribute:: count
    .. py:attribute:: ipdh
    """

    csdw_format = BitFormat('''
        u12 count
        p19
        u1 ipdh''')

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        fmt = 'u64 ipts'
        if self.ipdh:
            fmt += '\nu64 ipdh'
        fmt += '''
            u12 number
            u16 count
            u1 occurrence
            p3'''
        self.Message.FORMAT = BitFormat(fmt)

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: ipdh

            If present (see CSDW), contains the absolute time of the event.

        .. py:attribute:: number

            Event type number

        .. py:attribute:: count

            Number of events of this type as of the packet being written.

        .. py:attribute:: occurrence

            1 if event occurred during .RECORD mode.
        """
        def __repr__(self):
            return '<Recording event>'


class ComputerF3(packet.Packet):
    """Index Packet

    .. py:attribute:: count
    .. py:attribute:: ipdh
    .. py:attribute:: file_size_present
    .. py:attribute:: index_type

        * 0 - Root Index
        * 1 - Node Index

    .. py:attribute::file_size

        If enabled (see CSDW) indicates file size when packet was written.

    .. py:attribute::root_offset

        For root index, indicates offset to previous root index packet.
    """

    csdw_format = BitFormat('''
        u16 count
        p8
        u1 index_type
        u1 file_size_present
        u1 ipdh
        p5
    ''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: ipdh
        .. py:attribute:: offset

            Offset to packet from beginning of file.

        **Node Index Only**

        .. py:attribute:: channel_id
        .. py:attribute:: data_type
        """

        def __repr__(self):
            return '<%s Index>' % (
                'Node' if self.parent.index_type else 'Root')

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        if self.file_size_present and self.buffer:
            self.file_size, = bitstruct.unpack('u64<', self.buffer.read(8))

        fmt = 'u64 ipts'
        if self.ipdh:
            fmt += '\nu64 ipdh'

        if self.index_type == 0:
            fmt += '\nu64 offset'
        elif self.index_type == 1:
            fmt += '''
                u16 channel_id
                u8 data_type
                p8
                u64 offset'''

        self.Message.FORMAT = BitFormat(fmt)

        if self.index_type == 0 and self.buffer:
            pos = self.buffer.tell()
            self.buffer.seek(self.data_length - 8)
            self.root_offset, = bitstruct.unpack('u64<', self.buffer.read(8))
            self.buffer.seek(pos)

    def _read_messages(self):
        if self.buffer:
            offset = 36 if self.secondary_header else 24
            if self.file_size_present:
                offset += 8
            self.buffer.seek(offset + 4)
            self._messages = list(self)

    def _raw_body(self):
        raw = self.csdw_format.pack(self.__dict__)
        if self.file_size_present:
            raw += bitstruct.pack('u64<', self.file_size)
        raw += b''.join(bytes(m) for m in self._messages)
        if self.index_type == 0:
            raw += bitstruct.pack('u64<', self.root_offset)
        return raw




from .util import BitFormat
from . import packet


class PCMF1(packet.Packet):
    """IRIG 106 Chapter 4/8 PCM data.

    .. py:attribute:: sync_offset

        Offset into the major frame for the first data word in the packet. Not
        applicable for packed or throughput modes.

    .. py:attribute:: alignment

        Indicates 32 bit alignment if 1 or 16 bit by default.

    .. py:attribute:: throughput

        Indicates throughput mode (which would make most state flags
        unapplicable)

    .. py:attribute:: packed

        Indicates presence of packed data.

    .. py:attribute:: unpacked

        Indicates presense of unpacked data.

    .. py:attribute:: major_frame_status

        Combination major frame check flag and major frame lock

    .. py:attribute:: minor_frame_status

        Combination minor frame check flag and minor frame lock

    .. py:attribute:: minor_frame_indicator

        Indicates if the first data word in the packjet is the beginning of a
        minor frame.

    .. py:attribute:: major_frame_indicator

        Indicates if the first data word in the packet is the beginning of a
        major frame.

    .. py:attribute:: iph

        Indicates presence of Intra-Packet Header (IPTS and IPDH)
    """

    csdw_format = BitFormat('''
        u18 sync_offset
        u1 alignment
        u1 throughput
        u1 packed
        u1 unpacked
        p2
        u2 major_frame_status
        u2 minor_frame_status
        u1 minor_frame_indicator
        u1 major_frame_indicator
        u1 iph
        p1''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts

            Provided if IPH is true (see above).

        .. py:attribute:: lock_status
        """

        length = 12  # Two words sync, four data.

        def __repr__(self):
            return '<PCM Frame>'

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        # Throughput basically means we don't need to do anything.
        if self.throughput:
            self.Message = None

        else:
            fmt = '''
                u64 ipts
                p12
                u4 lock_status'''

            # Extra IPH word in 32 bit alignment.
            if self.iph and self.alignment:
                fmt += '\np16'

            self.Message.FORMAT = BitFormat(fmt)



#!/usr/bin/env python

from contextlib import suppress
from glob import glob
from setuptools import setup, Command
import os
import shutil
import sys

sys.path.insert(0, 'chapter10')
from version import version
sys.path.remove('chapter10')

cmdclass = {}
with suppress(ImportError):
    from sphinx.setup_command import BuildDoc
    cmdclass['build_docs'] = BuildDoc


class CleanCommand(Command):
    description = 'cleanup build and dist files'
    user_options = []

    CLEAN_FILES = '''
        build dist *.pyc *.tgz *.egg-info __pycache__ dependencies
        htmlcov docs/html docs/doctrees MANIFEST coverage.xml junit*.xml
    '''

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        here = os.path.abspath(os.path.dirname(__file__))
        for path_spec in self.CLEAN_FILES.split():
            abs_paths = glob(os.path.normpath(os.path.join(here, path_spec)))
            for path in abs_paths:
                print('removing %s' % os.path.relpath(path))
                if os.path.isdir(path):
                    shutil.rmtree(path, True)
                else:
                    with suppress(os.error):
                        os.remove(path)


cmdclass['clean'] = CleanCommand

setup(
    name='pychapter10',
    version=version,
    author='Micah Ferrill',
    author_email='ferrillm@avtest.com',
    description='A parser library for the IRIG 106 Chapter 10 data format.',
    long_description=open('README.rst').read(),
    url='https://github.com/atac/pychapter10',
    install_requires=['bitstruct>=8.11.0', 'bitstring>=3.1.7'],
    packages=['chapter10'],
    cmdclass=cmdclass,
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    command_options={
        'build_docs': {
            'build_dir': ('setup.py', 'docs',),
        }
    }
)




from .util import BitFormat
from . import packet


class DiscreteF1(packet.Packet):
    """Discrete Format 1

    .. py:attribute:: mode

        * 0 - data is recorded when the state changes
        * 1 - data is recorded at a timed interval

    .. py:attribute:: length

        Bit length of data or 0 (32 bits)
    """

    csdw_format = BitFormat('''
        u3 mode
        u5 length
        p24''')

    class Message(packet.Message):
        """.. py:attribute:: ipts"""

        length = 4
        FORMAT = BitFormat('u64 ipts')

        def __repr__(self):
            return '<Discrete Data %s bytes>' % len(self.data)




from .util import BitFormat
from . import packet


class MessageF0(packet.Packet):
    """
    .. py:attribute:: count

        Segment number of packet (for non-zero packet_type) or a count of \
messages contained in the packet.

    .. py:attribute:: packet_type

        * 0 - One or more complete messages
        * 1 - Beginning of a long message
        * 2 - Last part of a long message
        * 3 - Middle part of a long message
    """

    csdw_format = BitFormat('''
        u16 count
        u2 packet_type
        p14''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: length

            Message length (bytes)

        .. py:attribute:: subchannel
        .. py:attribute:: format_error
        .. py:attribute:: data_error
        """

        FORMAT = BitFormat('''
            u64 ipts
            u16 length
            u14 subchannel
            u1 format_error
            u1 data_error''')

        def __repr__(self):
            return '<Generic Message data %s bytes>' % len(self.data)




from .util import BitFormat
from . import packet


class MS1553F1(packet.Packet):
    """MIL-STD-1553B bus data

    .. py:attribute:: count

        Count of 1553 messages in the packet body.

    .. py:attribute:: time_tag_bits

        Indicates which bit of the message is time tagged:

        * 0 - Last bit of the last word of the message
        * 1 - First bit of the first word of the message
        * 2 - Last bit of the first (command) word of the message
    """

    csdw_format = BitFormat('''
        u24 count
        p6
        u2 time_tag_bits''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: le

            Length error

        .. py:attribute:: se

            Sync type error

        .. py:attribute:: we

            Invalid word error

        .. py:attribute:: bus

            Bus ID (A/B)

        .. py:attribute:: me

            Message error

        .. py:attribute:: fe

            Format error

        .. py:attribute:: timeout
        .. py:attribute:: gap_time

            GAP1 indicates time (in tenths of us) between the command or data \
word and the first (and only) status word. GAP2 measures time between the \
last data word and the second status word.

        .. py:attribute:: length

            Message length (bytes)
        """

        def __repr__(self):
            return '<1553F1 message %s words>' % (len(self.data) // 2)

        # Note: bitfields from status word are listed in different order than
        # shown in the standard. bitstruct doesn't allow for specifying bit
        # order across multiple fields.
        FORMAT = BitFormat('''
            u64 ipts

            p2
            u1 le
            u1 se
            u1 we
            p3

            p2
            u1 bus
            u1 me
            p1
            u1 fe
            u1 timeout
            p1

            u16 gap_time
            u16 length''')


class MS1553F2(packet.Packet):
    """16PP194 Weapons Bus Data

    **Note:** the specifics of the data word format (bus ID, etc.) are not \
implemented here (yet).

    .. py:attribute:: count
    """

    csdw_format = BitFormat('u32 count')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts
        .. py:attribute:: length

            IPDH length. Fixed at 24 bytes.

        .. py:attribute:: se

            Status error

        .. py:attribute:: ee

            Echo error

        .. py:attribute:: te

            Transaction error

        .. py:attribute:: re

            Master reset

        .. py:attribute:: tm

            Timeout
        """

        FORMAT = BitFormat('''
            u64 ipts
            u16 length
            u1 se
            p1
            u1 ee
            p3
            u1 te
            u1 re
            u1 tm
            p6''')

        def __repr__(self):
            return '<16PP194 Message %s words>' % (len(self.data) // 2)




from .util import BitFormat
from . import packet


class UARTF0(packet.Packet):
    """
    .. py:attribute:: iph
    """

    csdw_format = BitFormat('''<u32 iph''')

    class Message(packet.Message):
        """
        .. py:attribute:: ipts

            If enabled (see above), contains intra-packet timestamp

        .. py:attribute:: length

            Message length (bytes)

        .. py:attribute:: subchannel
        .. py:attribute:: parity_error
        """

        def __repr__(self):
            return '<UART data %s bytes>' % len(self.data)

    def __init__(self, *args, **kwargs):
        packet.Packet.__init__(self, *args, **kwargs)

        fmt = 'u64 ipts' if self.iph else ''
        fmt += '''
            u16 length
            u14 subchannel
            p1
            u1 parity_error'''
        self.Message.FORMAT = BitFormat(fmt)
